#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <flex.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "util.h"

#define BUFSIZE 1024
#define REPEAT (1024*1024)

int main() {
    int fd;
    char buf[BUFSIZE];
    struct timeval s1,f1, s2,f2, s3,f3, s4, f4;
    int ret, i;
    io_context_t ctx;
    struct io_event events[1];
    struct iocb cb, *cbp = &cb;
    struct timespec zerotime;
    zerotime.tv_sec = 0;
    zerotime.tv_nsec = 0;
    
    fd = open("/etc/services", O_RDONLY);
    if (fd <= 0) {
        perror("opening file");
        exit(-1);
    }
     
    memset(&ctx, 0 ,sizeof(ctx));
    if (io_setup(1024, &ctx) < 0) {
        perror ("setting up ioctx");
        exit (-1);
    }
   
    printf("synchronous reads...\n");
    gettimeofday (&s1, 0);
    for (i = 0; i < REPEAT; i++) {
        ret = read (fd, buf, sizeof(buf));
        if (ret < 0) {
            perror ("reading file");
            exit (-1);
        }
    }
    gettimeofday (&f1, 0);
   
    printf("aio (syscall)...\n");
    gettimeofday (&s2, 0);
    for (i = 0; i < REPEAT; i++) {
        io_prep_pread(cbp, fd, buf, sizeof(buf), 0);
        ret = io_submit (ctx, 1, &cbp);
        if (ret < 0) {
            perror ("submitting request");
            exit (-1);
        }
        if (_io_getevents(ctx, 1, 1, events , NULL) < 0) {
            perror ("getting events");
            exit (-1);
        }
    }            
    gettimeofday (&f2, 0);

    printf("aio (mapped event queue)...\n");
    gettimeofday (&s3, 0);
    for (i = 0; i < REPEAT; i++) {
        io_prep_pread(cbp, fd, buf, sizeof(buf), 0);
        ret = io_submit (ctx, 1, &cbp);
        if (ret < 0) {
            perror ("error submitting request");
            exit (-1);
        }
        if (io_getevents(ctx, 1, 1, events , NULL) < 0) {
            perror ("error getting events");
            exit (-1);
        }
    }            
    gettimeofday (&f3, 0);
   
    printf("aio (io_submit_try)...\n");
    gettimeofday (&s4, 0);
    for (i = 0; i < REPEAT; i++) {
        io_prep_pread(cbp, fd, buf, sizeof(buf), 0);
        memset (events, 0, sizeof (struct io_event));
        ret = io_submit_try (ctx, 1, &cbp, events);
        if (ret < 0) {
            perror ("error submitting request");
            exit (-1);
        }
	if (events[0].obj != cbp) {
	  printf("try_submit does not return event immediately, event.obj = %p\n", 
		 events[0].obj);
	  exit(-1);
	}
        if (events[0].res < 0) {
            perror ("aio request error");
            exit (-1);
        }
    }
    gettimeofday (&f4, 0);

    /* sanity check: make sure the event queue is empty */
    i = io_getevents(ctx, 1, 1, events, &zerotime);
    if (i != 0) {
      printf("Not zero events in queue after all tests: %d\n", i);
      exit(-1);
    }
            
    printf ("Total time: %g\n", duration (&s1, &f4));
    printf ("Blocking I/O: %g us, io_submit+_io_getevents(syscall): %g\n"
    	"io_submit+io_getevents(mmap): %g us, io_submit_try: %g us\n",
        duration (&s1, &f1) / REPEAT * 1000000,
        duration (&s2, &f2) / REPEAT * 1000000,
        duration (&s3, &f3) / REPEAT * 1000000,
        duration (&s4, &f4) / REPEAT * 1000000);
    
    return 0;
    
}
